
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <pspuser.h>
#include <pspdisplay.h>
#include <pspgu.h>
#include <pspgum.h>
#include <pspctrl.h>

#include "common.h"

#include "GU.h"
#include "VRAMManager.h"
#include "Texture.h"
#include "memtools.h"
#include "unicode.h"
#include "PlayList.h"
#include "LibImg.h"
#include "strtool.h"
#include "CFont.h"
#include "SystemSmallTexFont.h"
#include "sceHelper.h"
#include "euc2unicode.h"
#include "ImageCache.h"
#include "PlayTab.h"
#include "SimpleDialog.h"
#include "Lang.h"
#include "ProcState.h"

#include "Proc_FileList_BGImg.h"

#include "Proc_FileList_FileInfo.h"
#include "Proc_FileList_Folder.h"

static char LastBootFolderName[256]="ms0:/MUSIC";

static bool isDirectoryExists(const char *pPath)
{
  if(str_isEmpty(pPath)==true) return(false);
  
  SceUID uid=sceIoDopen(pPath);
  if(uid<0) return(false);
  sceIoDclose(uid);
  return(true);
}

static void Proc_PlayList_CallBack_ChangeState(void)
{
  Folder_RefreshExecutePlayIndex();
}

static void Proc_Start(EProcState ProcStateLast)
{
  BGImg_Load();
  
  Folder_Init();
  FileInfo_Init();
  
  if(isDirectoryExists(ProcState.LastPath)==false) StrCopy(ProcState.LastPath,BasePath);
  
  Folder_SetPath(ProcState.LastPath,NULL);
  
  PlayList_CallBack_ChangeState=Proc_PlayList_CallBack_ChangeState;
}

static void Proc_End(void)
{
  strcpy(ProcState.LastPath,Folder_GetCurrentPath());
  
  PlayList_CallBack_ChangeState=NULL;
  
  Folder_Free();
  FileInfo_Free();
  
  BGImg_Free();
}

static void Proc_KeyDown(u32 keys,u32 VSyncCount)
{
}

static void Proc_KeyUp(u32 keys,u32 VSyncCount)
{
}

static void Proc_KeyPress_ins_ExecuteDialog(void)
{
  CSimpleDialog *psd=new CSimpleDialog();
  
  psd->SetButtonsMask_OK(PSP_CTRL_SQUARE|PSP_CTRL_CIRCLE|PSP_CTRL_START|PSP_CTRL_SELECT);
  psd->SetButtonsMask_Cancel(PSP_CTRL_CROSS);
  
  psd->SetTitle(GetLangStr("System menu","システムメニュー"));
  
  psd->AddItem(GetLangStr("Settings","環境設定"));
  psd->AddItem(GetLangStr("Cancel","キャンセル"));
  
  psd->SetItemsIndex(ProcState.DialogIndex.SystemMenu);
  
  if(psd->ShowModal()==true){
    u32 idx=psd->GetItemsIndex();
    ProcState.DialogIndex.SystemMenu=idx;
    
    switch(idx){
      case 0: SetProcStateNext(EPS_Settings); break;
      case 1: break;
    }
  }
  
  if(psd!=NULL){
    delete psd; psd=NULL;
  }
}

static void Proc_KeyPress(u32 keys,u32 VSyncCount)
{
  if(keys&PSP_CTRL_UP) Folder_MoveCursorUp();
  if(keys&PSP_CTRL_DOWN) Folder_MoveCursorDown();
  if(keys&PSP_CTRL_LEFT) Folder_MoveCursorPageUp();
  if(keys&PSP_CTRL_RIGHT) Folder_MoveCursorPageDown();
  
  if(keys&PSP_CTRL_CIRCLE) Folder_ExecuteCurrentFile();
  
  if(keys&PSP_CTRL_CROSS){
    switch(ProcState.FileList.BButtonAssignment){
      case TProcState_FileList::EBBA_FolderUp: {
        Folder_MoveParentFolder();
      } break;
      case TProcState_FileList::EBBA_MusicStop: {
        if(PlayList_isOpened()==true) PlayList_Stop();
      } break;
      case TProcState_FileList::EBBA_Auto: {
        if(PlayList_isOpened()==true){
          PlayList_Stop();
          }else{
          Folder_MoveParentFolder();
        }
      } break;
    }
  }
  
  if(keys&PSP_CTRL_TRIANGLE) FileInfo_ToggleShow();
  
  if(keys&(PSP_CTRL_SQUARE|PSP_CTRL_START|PSP_CTRL_SELECT)) Proc_KeyPress_ins_ExecuteDialog();
}

static void Proc_KeysUpdate(TKeys *pk,u32 VSyncCount)
{
}

static void Proc_UpdateGU(u32 VSyncCount)
{
  PlayTab_ShowTitleBar();
  
  BGImg_VSyncUpdate(VSyncCount);
  BGImg_Draw();
  
  Folder_Update(VSyncCount);
  Folder_Draw();
  
  FileInfo_Draw(VSyncCount);
}

static bool Proc_UpdateBlank(void)
{
  return(false);
}

TProc_Interface* Proc_FileList_GetInterface(void)
{
  static TProc_Interface res={
    Proc_Start,
    Proc_End,
    Proc_KeyDown,
    Proc_KeyUp,
    Proc_KeyPress,
    Proc_KeysUpdate,
    Proc_UpdateGU,
    Proc_UpdateBlank,
  };
  return(&res);
}

