
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>

#include <pspuser.h>
#include <pspdisplay.h>
#include <pspgu.h>
#include <pspgum.h>
#include <pspctrl.h>

#include "common.h"
#include "GU.h"
#include "CFont.h"
#include "LibSnd.h"
#include "MemTools.h"
#include "SystemSmallTexFont.h"
#include "unicode.h"
#include "VRAMManager.h"
#include "Texture.h"
#include "LibImg.h"
#include "SndEff.h"
#include "ImageCache.h"
#include "strtool.h"
#include "ProcState.h"
#include "GlobalINI.h"
#include "Lang.h"
#include "PlayTab.h"

#include "Proc_Settings_BGImg.h"

#include "Proc_Settings_Menu.h"

typedef struct {
  typedef struct {
    u32 FontSize;
  } TFileList;
  TFileList FileList;
} TSettings;

static CMenu *pMenu;

static TTexture SE_GroupsFrameTex,SE_GroupsBarTex;
static TTexture SE_ItemsFrameTex,SE_ItemsBarTex;
static TTexture SE_TipLeftBtnTex,SE_TipRightBtnTex;

#include "Proc_Settings_MenuDraw.h"

static void Proc_Start(EProcState ProcStateLast)
{
  BGImg_Load();
  
  KeyControl_AnalogPadToDigialPad=true;
  
  pMenu=new CMenu(&ProcState.GroupsIndex);
  
  CMenu *pm=pMenu;
  
  static u8 value=0;
  
  {
    TProcState_Global *ps=&ProcState.Global;
    CMenuGroup *pmg=pm->CreateGroup(&ps->ItemsIndex,GetLangStr("Global","全体設定"));
    {
      CMenuGroupItem *pmgi=pmg->CreateItem(GetLangStr("No setting item","設定項目がありません。"),"",(u8*)&value);
      pmgi->CreateSet(GetLangStr("X",""),GetLangStr("",""),0);
      pmgi->SetDefault();
    }
    pmg->SetDefault();
  }
  
  {
    TProcState_FileList *ps=&ProcState.FileList;
    CMenuGroup *pmg=pm->CreateGroup(&ps->ItemsIndex,GetLangStr("FileList",""));
    {
      CMenuGroupItem *pmgi=pmg->CreateItem(GetLangStr("Filename font size","ファイル名サイズ"),"",(u8*)&ps->FilenameFontSize);
      pmgi->CreateSet(GetLangStr("a12A","A12あ"),GetLangStr("Very small","とても小さい"),12,12);
      pmgi->CreateSet(GetLangStr("a14A","A14あ"),GetLangStr("Small","小さい"),14,14);
      pmgi->CreateSet(GetLangStr("a16A","A16あ"),GetLangStr("Normal","標準"),16,16);
      pmgi->CreateSet(GetLangStr("a20A","A20あ"),GetLangStr("Large","大きい"),20,20);
      pmgi->SetDefault();
    }
    {
      CMenuGroupItem *pmgi=pmg->CreateItem(GetLangStr("Show file icon","ファイルアイコン"),"",(u8*)&ps->ShowFileIcon);
      pmgi->CreateSet(GetLangStr("OFF",""),GetLangStr("Filename only.","ファイル名のみ表示"),0);
      pmgi->CreateSet(GetLangStr("ON",""),GetLangStr("Icon and filename.","アイコンとファイル名を表示"),1);
      pmgi->SetDefault();
    }
    {
      CMenuGroupItem *pmgi=pmg->CreateItem(GetLangStr("Number of lines","ファイル名行数"),"",(u8*)&ps->NumberOfLines);
      pmgi->CreateSet(GetLangStr("1",""),GetLangStr("A long file name cuts.","長いファイル名は切り捨てる"),1);
      pmgi->CreateSet(GetLangStr("2",""),GetLangStr("A long file name is divided.","長いファイル名は二行に分ける"),2);
      pmgi->SetDefault();
    }
    {
      CMenuGroupItem *pmgi=pmg->CreateItem(GetLangStr("B button assignment","Bボタン割り当て"),"",(u8*)&ps->BButtonAssignment);
      pmgi->CreateSet(GetLangStr("Folder-up","フォルダ戻り"),GetLangStr("Exit folder.","一つ前のフォルダに戻る"),TProcState_FileList::EBBA_FolderUp);
      pmgi->CreateSet(GetLangStr("Music-stop","音楽停止"),GetLangStr("Music stop.","再生中の音楽を停止する"),TProcState_FileList::EBBA_MusicStop);
      pmgi->CreateSet(GetLangStr("Auto","自動"),GetLangStr("Exit folder or music stop.","再生中なら音楽停止。又はフォルダから出る"),TProcState_FileList::EBBA_Auto);
      pmgi->SetDefault();
    }
    pmg->SetDefault();
  }
  
  {
    TProcState_PlayTab *ps=&ProcState.PlayTab;
    CMenuGroup *pmg=pm->CreateGroup(&ps->ItemsIndex,GetLangStr("PlayTab",""));
    {
      CMenuGroupItem *pmgi=pmg->CreateItem(GetLangStr("Filename font size (Req.reboot)","ファイル名サイズ（要再起動）"),"",(u8*)&ps->FilenameFontSize);
      pmgi->CreateSet(GetLangStr("a12A","A12あ"),GetLangStr("Very small","とても小さい"),12,12);
      pmgi->CreateSet(GetLangStr("a14A","A14あ"),GetLangStr("Small","小さい"),14,14);
      pmgi->CreateSet(GetLangStr("a16A","A16あ"),GetLangStr("Normal","標準"),16,16);
      pmgi->CreateSet(GetLangStr("a20A","A20あ"),GetLangStr("Large","大きい"),20,20);
      pmgi->CreateSet(GetLangStr("a24A","A24あ"),GetLangStr("Very large","とても大きい"),24,24);
      pmgi->SetDefault();
    }
    {
      CMenuGroupItem *pmgi=pmg->CreateItem(GetLangStr("Number of lines","ファイル名行数"),"",(u8*)&ps->NumberOfLines);
      pmgi->CreateSet(GetLangStr("1",""),GetLangStr("Current + Next","現在曲"),1);
      pmgi->CreateSet(GetLangStr("2",""),GetLangStr("Current + Next","現在曲＋次曲"),2);
      pmgi->CreateSet(GetLangStr("3",""),GetLangStr("Back + Current + Next","前曲＋現在曲＋次曲"),3);
      pmgi->CreateSet(GetLangStr("4",""),GetLangStr("Back + Current + Next + Next","前曲＋現在曲＋次曲＋次曲"),4);
      pmgi->SetDefault();
    }
    {
      CMenuGroupItem *pmgi=pmg->CreateItem(GetLangStr("Music info font size (Req.reboot)","曲情報サイズ（要再起動）"),"",(u8*)&ps->MusicInfoFontSize);
      pmgi->CreateSet(GetLangStr("a12A","A12あ"),GetLangStr("Very small","とても小さい"),12,12);
      pmgi->CreateSet(GetLangStr("a14A","A14あ"),GetLangStr("Small","小さい"),14,14);
      pmgi->CreateSet(GetLangStr("a16A","A16あ"),GetLangStr("Normal","標準"),16,16);
      pmgi->CreateSet(GetLangStr("a20A","A20あ"),GetLangStr("Large","大きい"),20,20);
      pmgi->CreateSet(GetLangStr("a24A","A24あ"),GetLangStr("Very large","とても大きい"),24,24);
      pmgi->SetDefault();
    }
    {
      CMenuGroupItem *pmgi=pmg->CreateItem(GetLangStr("Play list mode","プレイリスト演奏モード"),"",(u8*)&ps->PlayListMode);
      pmgi->CreateSet(GetLangStr("Normal","順再生"),GetLangStr("",""),TProcState_PlayTab::EPLM_Normal);
      pmgi->SetDefault();
    }
    pmg->SetDefault();
  }
  
  {
    TProcState_Image *ps=&ProcState.Image;
    CMenuGroup *pmg=pm->CreateGroup(&ps->ItemsIndex,GetLangStr("Image",""));
    {
      CMenuGroupItem *pmgi=pmg->CreateItem(GetLangStr("Auto jpeg fitting.","JPEGファイル自動縮小"),"",(u8*)&ps->AutoJpegFitting);
      pmgi->CreateSet(GetLangStr("ON",""),GetLangStr("An extremely big JPEG file is reduced.","大きすぎる画像は自動調整する"),1);
      pmgi->SetDefault();
    }
    pmg->SetDefault();
  }
  
  {
    TProcState_Text *ps=&ProcState.Text;
    CMenuGroup *pmg=pm->CreateGroup(&ps->ItemsIndex,GetLangStr("Text",""));
    {
      CMenuGroupItem *pmgi=pmg->CreateItem(GetLangStr("No setting item","設定項目がありません。"),"",(u8*)&value);
      pmgi->CreateSet(GetLangStr("X",""),GetLangStr("",""),0);
      pmgi->SetDefault();
    }
    pmg->SetDefault();
  }
  
  pm->SetDefault();
  
  pm->ShowTreeInfo();
  
  Texture_CreateFromFile(true,EVMM_Process,&SE_GroupsFrameTex,ETF_RGBA8888,Resources_SettingsPath "/SE_GroupsFrame.png");
  Texture_CreateFromFile(false,EVMM_Process,&SE_GroupsBarTex,ETF_RGBA8888,Resources_SettingsPath "/SE_GroupsBar.png");
  Texture_CreateFromFile(true,EVMM_Process,&SE_ItemsFrameTex,ETF_RGBA8888,Resources_SettingsPath "/SE_ItemsFrame.png");
  Texture_CreateFromFile(false,EVMM_Process,&SE_ItemsBarTex,ETF_RGBA8888,Resources_SettingsPath "/SE_ItemsBar.png");
  Texture_CreateFromFile(false,EVMM_Process,&SE_TipLeftBtnTex,ETF_RGBA8888,Resources_SettingsPath "/SE_TipLeftBtn.png");
  Texture_CreateFromFile(false,EVMM_Process,&SE_TipRightBtnTex,ETF_RGBA8888,Resources_SettingsPath "/SE_TipRightBtn.png");
  
  DrawMenu_Init();
}

static void Proc_End(void)
{
  BGImg_Free();
  
  KeyControl_AnalogPadToDigialPad=false;
}

static void Proc_KeyDown(u32 keys,u32 VSyncCount)
{
}

static void Proc_KeyUp(u32 keys,u32 VSyncCount)
{
}

static void ExitToFileList(bool StateSave)
{
  SndEff_Play(ESE_Success);
  
  if(StateSave==true){
    }else{
    PlayTab_Refresh();
  }
  
  SetProcStateNext(EPS_FileList);
}

static void Proc_KeyPress(u32 keys,u32 VSyncCount)
{
  CMenu *pm=pMenu;
  
  CMenu::ESelectState SelectState=pm->GetSelectState();
  
  if(keys&(PSP_CTRL_START|PSP_CTRL_SELECT)) ExitToFileList(false);
  
  if(keys&PSP_CTRL_CROSS){
    switch(SelectState){
      case CMenu::ESS_Group: {
        ExitToFileList(true);
      } break;
      case CMenu::ESS_Item: {
        SndEff_Play(ESE_MoveOut);
        pm->SetSelectState(CMenu::ESS_Group);
      }
    }
  }
  
  if(keys&PSP_CTRL_CIRCLE){
    switch(SelectState){
      case CMenu::ESS_Group: {
        SndEff_Play(ESE_MoveIn);
        pm->SetSelectState(CMenu::ESS_Item);
      } break;
      case CMenu::ESS_Item: {
        ExitToFileList(true);
      } break;
    }
  }
  
  if(keys&(PSP_CTRL_CIRCLE|PSP_CTRL_CROSS|PSP_CTRL_LEFT|PSP_CTRL_RIGHT|PSP_CTRL_UP|PSP_CTRL_DOWN)){
    MenuGlobal_KeyPress();
  }
  
  switch(SelectState){
    case CMenu::ESS_Group: {
      {
        s32 v=0;
        if(keys&PSP_CTRL_UP) v=-1;
        if(keys&PSP_CTRL_DOWN) v=+1;
        if(v!=0) MenuGroup_MoveCursor(v);
      }
    } break;
    case CMenu::ESS_Item: {
      {
        s32 v=0;
        if(keys&PSP_CTRL_UP) v=-1;
        if(keys&PSP_CTRL_DOWN) v=+1;
        if(v!=0) MenuItem_MoveCursor(v);
      }
      {
        s32 v=0;
        if(keys&PSP_CTRL_LEFT) v=-1;
        if(keys&PSP_CTRL_RIGHT) v=+1;
        if(v!=0){
          u32 last=ProcState.PlayTab.NumberOfLines;
          MenuItem_ChangeValue(v,true);
          if(last!=ProcState.PlayTab.NumberOfLines) PlayTab_Refresh();
        }
      }
    }
  }
}

static void Proc_KeysUpdate(TKeys *pk,u32 VSyncCount)
{
}

static void Proc_UpdateGU(u32 VSyncCount)
{
  BGImg_VSyncUpdate(VSyncCount);
  BGImg_Draw();
  
  Texture_GU_Start();
  
  CMenu *pm=pMenu;
  
  {
    float f=pm->GetFadeValue();
    pm->UpdateVSync(VSyncCount);
    f=f-pm->GetFadeValue();
    BGImg_Particle_SetAdd(-f*64,0);
  }
  
  CMenu::ESelectState SelectState=pm->GetSelectState();
  
  CMenuGroup *pmg=pm->GetCurrentGroup();
  DrawMenu_Item(VSyncCount,SelectState,pmg);
  
  DrawMenu_Group(VSyncCount,SelectState,pm);
  
  Texture_GU_End();
}

static bool Proc_UpdateBlank(void)
{
  return(false);
}

TProc_Interface* Proc_Settings_GetInterface(void)
{
  static TProc_Interface res={
    Proc_Start,
    Proc_End,
    Proc_KeyDown,
    Proc_KeyUp,
    Proc_KeyPress,
    Proc_KeysUpdate,
    Proc_UpdateGU,
    Proc_UpdateBlank,
  };
  return(&res);
}


