
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pspaudiolib.h>
#include <pspaudio.h>
#include <pspdisplay.h>

#include "common.h"
#include "memtools.h"

#include "LibSnd.h"

#include "SndEff.h"

static const bool UseSE=false;

typedef struct {
  const char *pfn;
  u32 SamplesCount;
  u32 *pSamples;
} TWave;

static TWave Waves[ESE_Count];

typedef struct {
  volatile u32 *pSamples;
  volatile u32 SamplesCount;
} TPlayData;

static const u32 PlayDatasCount=4;
static TPlayData PlayDatas[PlayDatasCount];

static void Wave_Init(void)
{
  for(u32 idx=0;idx<PlayDatasCount;idx++){
    TPlayData *ppd=&PlayDatas[idx];
    ppd->pSamples=NULL;
    ppd->SamplesCount=0;
  }
}

static void Wave_Stop(void)
{
  Wave_Init();
}

static u32 Wave_GetFreeCh(void)
{
  u32 SmallSamplesCount=0xffffffff;
  u32 ch=(u32)-1;
  for(u32 idx=0;idx<PlayDatasCount;idx++){
    TPlayData *ppd=&PlayDatas[idx];
    if(ppd->pSamples==NULL) return(idx);
    if(ppd->SamplesCount<SmallSamplesCount){
      SmallSamplesCount=ppd->SamplesCount;
      ch=idx;
    }
  }
  
  if(ch==(u32)-1){
    printf("Can not found free channel for SE.\n");
    SystemHalt();
  }
  
  return(ch);
}

static void Wave_Play(TWave *pw)
{
  u32 ch=Wave_GetFreeCh();
  
  TPlayData *ppd=&PlayDatas[ch];
  
  ppd->pSamples=pw->pSamples;
  ppd->SamplesCount=pw->SamplesCount;
}

static void Wave_Load(TWave *pw,s32 vol8,const char *pfn)
{
  pw->pfn=pfn;
  pw->SamplesCount=0;
  pw->pSamples=NULL;
  
  if(LibSnd_Start(false,pfn,0)==false){
    printf("Can not start SE decoder. [%s]\n",pfn);
    SystemHalt();
  }
  
  const TLibSndConst_State *pState=LibSnd_GetState();
  const u32 SamplesParFrame=pState->SamplesPerFlame;
  
  while(1){
    pw->pSamples=(u32*)saferealloc(pw->pSamples,sizeof(u32)*(pw->SamplesCount+SamplesParFrame));
    u32 SamplesCount=LibSnd_InternalDecode(&pw->pSamples[pw->SamplesCount]);
    if(SamplesCount==0) break;
    pw->SamplesCount+=SamplesCount;
  }
  
  LibSnd_Close();
  
  for(u32 idx=0;idx<pw->SamplesCount;idx++){
    u32 smp=pw->pSamples[idx];
    s16 l=smp&0xffff,r=smp>>16;
    l=(l*vol8)/0x100;
    r=(r*vol8)/0x100;
    smp=(l&0xffff)|(r<<16);
    pw->pSamples[idx]=smp;
  }
  
  printf("%d samples decoded.\n",pw->SamplesCount);
}

static void Wave_Free(TWave *pw)
{
  pw->pfn=NULL;
  pw->SamplesCount=0;
  if(pw->pSamples!=NULL){
    safefree(pw->pSamples); pw->pSamples=NULL;
  }
}

static void audioSECallback(void* buf, unsigned int length, void *userdata)
{
  u32 buflength=length;
  u32 *pbuf32=(u32*)buf;
  
  TPlayData pds[PlayDatasCount];
  
  for(u32 idx=0;idx<PlayDatasCount;idx++){
    pds[idx]=PlayDatas[idx];
  }
  
  for(u32 idx=0;idx<buflength;idx++){
    pbuf32[idx]=0;
  }
  
  for(u32 idx=0;idx<PlayDatasCount;idx++){
    TPlayData *ppd=&pds[idx];
    if(ppd->pSamples!=NULL){
      for(u32 idx=0;idx<buflength;idx++){
        u32 srcsmp=pbuf32[idx];
        s32 l=(s16)(srcsmp&0xffff),r=(s16)(srcsmp>>16);
        u32 smp=*ppd->pSamples++;
        l+=(s16)(smp&0xffff);
        if(l<-32768) l=-32768;
        if(32767<l) l=32767;
        r+=(s16)(smp>>16);
        if(r<-32768) r=-32768;
        if(32767<r) r=32767;
        pbuf32[idx]=(l&0xffff)|(r<<16);
        ppd->SamplesCount--;
        if(ppd->SamplesCount==0){
          ppd->pSamples=NULL;
          break;
        }
      }
    }
  }
  
  for(u32 idx=0;idx<PlayDatasCount;idx++){
    PlayDatas[idx]=pds[idx];
  }
}

void SndEff_Init(void)
{
  if(UseSE==false) return;
  
  Wave_Init();
  
  pspAudioSetVolume(1, PSP_VOLUME_MAX,PSP_VOLUME_MAX);
  pspAudioSetChannelCallback(1, audioSECallback, NULL);
  
  Wave_Load(&Waves[ESE_Boot],0x40,Resources_SEPath "/Test_Android31_Media_Audio/Boot.tta");
  Wave_Play(&Waves[ESE_Boot]);
  
  Wave_Load(&Waves[ESE_Warrning],0x80,Resources_SEPath "/Test_Android31_Media_Audio/Warrning.tta");
  Wave_Load(&Waves[ESE_Dialog],0x80,Resources_SEPath "/Test_Android31_Media_Audio/Dialog.tta");
  Wave_Load(&Waves[ESE_Success],0x20,Resources_SEPath "/Test_Android31_Media_Audio/Success.tta");
  Wave_Load(&Waves[ESE_Error],0x80,Resources_SEPath "/Test_Android31_Media_Audio/Error.tta");
  Wave_Load(&Waves[ESE_Open],0x80,Resources_SEPath "/Test_Android31_Media_Audio/Open.tta");
  Wave_Load(&Waves[ESE_MoveIn],0xff,Resources_SEPath "/Test_Android31_Media_Audio/MoveIn.tta");
  Wave_Load(&Waves[ESE_MoveOut],0xff,Resources_SEPath "/Test_Android31_Media_Audio/MoveOut.tta");
  Wave_Load(&Waves[ESE_MovePage],0x80,Resources_SEPath "/Test_Android31_Media_Audio/MovePage.tta");
  Wave_Load(&Waves[ESE_KeyPress],0xc0,Resources_SEPath "/Test_Android31_Media_Audio/KeyPress.tta");
  Wave_Load(&Waves[ESE_ShortNoise],0x40,Resources_SEPath "/Test_Android31_Media_Audio/ShortNoise.tta");
}

void SndEff_Free_BootOnly(void)
{
  if(UseSE==false) return;
  
  TWave *pw=&Waves[ESE_Boot];
  if(pw->pSamples==NULL) return;
  
  for(u32 idx=0;idx<PlayDatasCount;idx++){
    TPlayData *ppd=&PlayDatas[idx];
    if(ppd->pSamples!=NULL) return;
  }
  
  Wave_Free(&Waves[ESE_Boot]);
}

void SndEff_Free(void)
{
  if(UseSE==false) return;
  
  Wave_Stop();
}

void SndEff_Play(ESE ese)
{
  if(UseSE==false) return;
  
  TWave *pw=&Waves[ese];
  if(pw->pSamples==NULL){
    printf("Not loaded SE. (%d)\n",ese);
    SystemHalt();
  }
  
  Wave_Play(pw);
}

