
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pspuser.h>

#include "euc2unicode.h"
#include "strtool.h"

#include "LibSnd_Const_Global.h"
#include "LibSnd_Const_Internal.h"

static TLibSndConst_State LibSndInt_State;

// ------------------------------------------------------------------------------------

#include "game-music-emu-0.5.5/Music_Emu.h"

static u32 TrackNum;

static u8 *pDataBuf;
static u32 DataBufSize;

static Music_Emu *pGME;
static gme_info_t *pGMEInfo;

// ------------------------------------------------------------------------------------

// ------------------------------------------------------------------------------------

static void LibSndInt_ShowLicense(void)
{
  printf("Game_Music_Emu 0.5.5\n");
  printf("--------------------\n");
  printf("Author : Shay Green <gblargg@gmail.com>\n");
  printf("Website: http://www.slack.net/~ant/libs/\n");
  printf("Forum  : http://groups.google.com/group/blargg-sound-libs\n");
  printf("License: GNU Lesser General Public License (LGPL)\n");
  printf("\n");
  printf("LibSndGME by Moonlight.\n");
  printf("\n");
}

static bool LibSndInt_Open(const char *pFilename,const u32 _TrackNum)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  {
    FILE *fp=fopen(pFilename,"r");
    if(fp==NULL){
      snprintf(pState->ErrorMessage,256,"File not found. [%s]",pFilename);
      return(false);
    }
    
    fseek(fp,0,SEEK_END);
    DataBufSize=ftell(fp);
    fseek(fp,0,SEEK_SET);
    pDataBuf=(u8*)malloc(DataBufSize);
    fread(pDataBuf,DataBufSize,1,fp);
    
    if(fp!=NULL){
      fclose(fp); fp=NULL;
    }
  }
  
  TrackNum=_TrackNum;
  
  u32 SampleRate=44100;
  u32 SamplePerFlame=512;
  
  gme_err_t err=gme_open_data(pDataBuf,DataBufSize,&pGME,SampleRate);
  if(err!=NULL){
    snprintf(pState->ErrorMessage,256,"%s",err);
    if(pDataBuf!=NULL){
      free(pDataBuf); pDataBuf=NULL;
    }
    return(false);
  }
  
  gme_set_stereo_depth(pGME,1);
  
  err=gme_start_track(pGME,TrackNum);
  if(err!=NULL){
    snprintf(pState->ErrorMessage,256,"%s",err);
    if(pGME!=NULL){
      gme_delete(pGME); pGME=NULL;
    }
    if(pDataBuf!=NULL){
      free(pDataBuf); pDataBuf=NULL;
    }
    return(false);
  }
  
  {
    gme_err_t err=gme_track_info(pGME,&pGMEInfo,TrackNum);
    if(err!=NULL){
      snprintf(pState->ErrorMessage,256,"%s",err);
      if(pGME!=NULL){
        gme_delete(pGME); pGME=NULL;
      }
      if(pDataBuf!=NULL){
        free(pDataBuf); pDataBuf=NULL;
      }
      return(false);
    }
  }
  
  u32 LengthSec=0;
  if(0<pGMEInfo->length) LengthSec=pGMEInfo->length;
  if(LengthSec==0) LengthSec=90;
  
  {
    
    char msg[1024];
    {
      char *pmsg=msg;
      pmsg[0]=0;
      if(pGMEInfo->game!=NULL) pmsg+=snprintf(pmsg,128,"Game:%s",pGMEInfo->game);
      if(pGMEInfo->song!=NULL) pmsg+=snprintf(pmsg,128,"/ Song:%s",pGMEInfo->song);
      if(pGMEInfo->author!=NULL) pmsg+=snprintf(pmsg,128,"/ Author:%s",pGMEInfo->author);
      if(pGMEInfo->copyright!=NULL) pmsg+=snprintf(pmsg,128,"/ Copyright:%s",pGMEInfo->copyright);
      if(pGMEInfo->comment!=NULL) pmsg+=snprintf(pmsg,128,"/ Comment:%s",pGMEInfo->comment);
      if(pGMEInfo->dumper!=NULL) pmsg+=snprintf(pmsg,128,"/ Dumper:%s",pGMEInfo->dumper);
      if(pGMEInfo->system!=NULL) pmsg+=snprintf(pmsg,128,"/ System:%s",pGMEInfo->system);
    }
    pState->pTitleW=EUC2Unicode_Convert(msg);
    
  }
  
  pState->SampleRate=SampleRate;
  pState->SamplesPerFlame=SamplePerFlame;
  pState->isEnd=false;
  pState->CurrentSec=0;
  pState->TotalSec=(float)LengthSec;
  
  LibSnd_DebugOut("GME decoder initialized.\n");
  
  return(true);
}

static u32 LibSndInt_Update(u32 *pBufLR);

static void LibSndInt_Seek(float sec)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  if(sec<pState->CurrentSec){
    gme_err_t err=gme_start_track(pGME,TrackNum);
    if(err!=NULL){
      printf(pState->ErrorMessage,256,"%s",err);
      return;
    }
    pState->CurrentSec=0;
  }
  
  while(pState->CurrentSec<sec){
    if(LibSndInt_Update(NULL)==0) break;
  }
}

static s32 lastl=0,lastr=0;

static u32 LibSndInt_Update(u32 *pBufLR)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  if(gme_track_ended(pGME)==true) pState->isEnd=true;
  
  if(pState->isEnd==true) return(0);
  
  if(pBufLR==NULL){
    gme_set_stereo_depth(pGME,0);
    pGME->skip(pState->SamplesPerFlame*2);
    gme_set_stereo_depth(pGME,1);
    }else{
    s16 *pBufLR16=(s16*)pBufLR;
    gme_err_t err=gme_play(pGME,pState->SamplesPerFlame*2,pBufLR16);
    if(err!=NULL){
      snprintf(pState->ErrorMessage,256,"%s",err);
      return(0);
    }
    s32 tmpl=lastl,tmpr=lastr;
    for(u32 idx=0;idx<pState->SamplesPerFlame;idx++){
      s32 l=pBufLR16[0],r=pBufLR16[1];
      
if(0){
      // SimpleLPF Normal
      l=(l+tmpl)/2; r=(r+tmpr)/2;
      tmpl=l; tmpr=r;
}
      
      pBufLR16[0]=l; pBufLR16[1]=r;
      pBufLR+=2;
    }
    lastl=tmpl; lastr=tmpr;
  }
  
  pState->CurrentSec+=(float)pState->SamplesPerFlame/pState->SampleRate;
  
  return(pState->SamplesPerFlame);
}

static void LibSndInt_Close(void)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  if(pGMEInfo!=NULL){
    gme_free_info(pGMEInfo); pGMEInfo=NULL;
  }
  
  if(pGME!=NULL){
    gme_delete(pGME); pGME=NULL;
  }
}

extern u32 LibSndInt_Helper_GetTracksCount(FILE *fp);
u32 LibSndInt_Helper_GetTracksCount(FILE *fp)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  u8 *pDataBuf;
  u32 DataBufSize;
  
  fseek(fp,0,SEEK_END);
  DataBufSize=ftell(fp);
  fseek(fp,0,SEEK_SET);
  pDataBuf=(u8*)malloc(DataBufSize);
  fread(pDataBuf,DataBufSize,1,fp);
  
  Music_Emu *pGME;
  
  gme_err_t err=gme_open_data(pDataBuf,DataBufSize,&pGME,44100);
  if(err!=NULL){
    snprintf(pState->ErrorMessage,256,"%s",err);
    if(pDataBuf!=NULL){
      free(pDataBuf); pDataBuf=NULL;
    }
    return(0);
  }
  
  u32 TracksCount=gme_track_count(pGME);
  
  if(pGME!=NULL){
    gme_delete(pGME); pGME=NULL;
  }
  
  if(pDataBuf!=NULL){
    free(pDataBuf); pDataBuf=NULL;
  }
  
  return(TracksCount);
}

static const char* GetInfoStr(gme_info_t *pGMEInfo,u32 idx)
{
  const u32 strlen=256;
  static char str[strlen+1];
  str[0]=0;
  
  {
    if(idx==0){
      snprintf(str,strlen,"Track: %d/%d.",1+TrackNum,gme_track_count(pGME));
      return(str);
    }
    idx--;
  }
  
  {
    if(idx==0){
      snprintf(str,strlen,"Total: %d, Intro: %d, Loop: %d.",pGMEInfo->length,pGMEInfo->intro_length,pGMEInfo->loop_length);
      return(str);
    }
    idx--;
  }
  
#define get(Name,value) { \
  if(str_isEmpty(value)==false){ \
    if(idx==0){ \
      snprintf(str,strlen,"%s: %s",Name,value); \
      return(str); \
    } \
    idx--; \
  } \
}
#define getnum(Name,num,value) { \
  if(str_isEmpty(value)==false){ \
    if(idx==0){ \
      snprintf(str,strlen,"%s%d: %s",Name,num,value); \
      return(str); \
    } \
    idx--; \
  } \
}
  
  get("Game",pGMEInfo->game);
  get("Song",pGMEInfo->song);
  get("Author",pGMEInfo->author);
  get("CopyRight",pGMEInfo->copyright);
  get("Comment",pGMEInfo->comment);
  get("Dumper",pGMEInfo->dumper);
  get("System",pGMEInfo->system);
  
  for(u32 vidx=0;vidx<gme_voice_count(pGME);vidx++){
    getnum("Trk",vidx,gme_voice_name(pGME,vidx));
  }
  
  return(NULL);

#undef get
#undef getnum
}

static int LibSndInt_GetInfoCount(void)
{
  u32 cnt=0;
  
  while(1){
    const char *pstr=GetInfoStr(pGMEInfo,cnt);
    if(str_isEmpty(pstr)==true) break;
    cnt++;
  }
  
  return(cnt);
}

static bool LibSndInt_GetInfoStrUTF8(int idx,char *str,int len)
{
  return(false);
}

static bool LibSndInt_GetInfoStrW(int idx,wchar *str,int len)
{
  const char *pstr=GetInfoStr(pGMEInfo,idx);
  if(str_isEmpty(pstr)==true) return(false);
  
  wchar *pstrw=EUC2Unicode_Convert(pstr);
  Unicode_CopyNum(str,pstrw,len);
  if(pstrw!=NULL){
    safefree(pstrw); pstrw=NULL;
  }
  
  return(true);
}

TLibSnd_Interface* LibSndGME_GetInterface(void)
{
  static TLibSnd_Interface res={
    LibSndInt_ShowLicense,
    LibSndInt_Open,
    LibSndInt_Seek,
    LibSndInt_Update,
    LibSndInt_Close,
    LibSndInt_GetInfoCount,
    LibSndInt_GetInfoStrUTF8,
    LibSndInt_GetInfoStrW,
    &LibSndInt_State,
  };
  return(&res);
}

